/* CmdScan 1.0 -- Command line scanning utility
 * Uses the SANE library.
 * Copyright (C) 1996 Andreas Beck
 *
 * You can contact me at becka@sunserver1.rz.uni-duesseldorf.de
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "sane.h"

/* Declare local functions. */

static SANE_Handle scan_handle;
static int verbose=0;
static int help=0;

static void scan_it(void)
{
  SANE_Parameters parm;
  int len;
  char buffer[38400];
  
  if (sane_start(scan_handle)!=0) {fprintf(stderr,"sane_start failed.\n");return;}
  sane_get_parameters(scan_handle,&parm);
  
  switch(parm.format)
  { case SANE_FRAME_RGB:
          printf("P6\n# SANE data follows\n%d %d\n255\n",parm.pixels_per_line,parm.lines);
          break;
    default:
          printf("P5\n# SANE data follows\n%d %d\n255\n",parm.pixels_per_line,parm.lines);
          break;
  }    
  while(sane_read(scan_handle,buffer,parm.bytes_per_line,&len)==SANE_STATUS_GOOD)
  { fwrite(buffer,1,len,stdout); }

  sane_cancel(scan_handle);
}

void errout(char *p)
{ fprintf(stderr,"%s\n",p);exit(1); }

char *getoption(int argc, char **argv, char * option)
{ int x;
  for(x=1;x<argc;x++)
     if (*argv[x]=='-'&&strncmp(argv[x]+1,option,4)==0) return argv[x+1];
  return NULL;
}

int main (int argc, char **argv)
{ int hlp;
  int x;
  char *gopt;
  const SANE_Option_Descriptor *opt;
  const SANE_Device * * device_list;
  
  if (getoption(argc,argv,"?")) 
  { help=1; fprintf(stderr,"CmdScan SANE commandline-scanning-utility:\n"); }

  if (getoption(argc,argv,"v")) 
  { verbose=1; fprintf(stderr,"CmdScan SANE commandline-scanning-utility verbose mode:\n"); }

  sane_init(NULL);

  /* Open the first scanner device. FIXME. Should make an option for that. */

  sane_get_devices(&device_list);

  if (sane_open(device_list[0]->name,&scan_handle)==SANE_STATUS_GOOD)
  { 
    for(x=0;(opt=sane_get_option_descriptor(scan_handle,x))!=NULL;x++)
    { switch(opt->type)
      { 
        case SANE_TYPE_GROUP:
        	if (help) fprintf(stderr,"\n%s\n",opt->name);
        	break;

        case SANE_TYPE_BOOL:
        	if (help) fprintf(stderr,"-%-10.10s [yn01]   %s\n",
        				opt->name,opt->desc);
        	if ((gopt=getoption(argc,argv,opt->name))==NULL) break;
        	switch(*gopt)
        	{ case '0':
        	  case 'n':
        	  case 'N':hlp=0;break;
        	  case '1':
        	  case 'y':
        	  case 'Y':hlp=1;break;
        	  default :errout("Invalid parameter for toggle option.");
        	} 
	    	if (sane_control_option(scan_handle,x,SANE_ACTION_SET_VALUE,&hlp,NULL)<0) 
	    		errout("Option not O.K.");
        	break;
        case SANE_TYPE_INT:
	        if (help) fprintf(stderr,"-%-10.10s          %s\n",
        		opt->name,opt->desc);
        	if ((gopt=getoption(argc,argv,opt->name))==NULL) break;
	    	hlp=(int) atof(gopt);
	    	if (verbose) fprintf(stderr,"%d\n",hlp);
	    	if (sane_control_option(scan_handle,x,SANE_ACTION_SET_VALUE,&hlp,NULL)<0) 
		  errout("Option not O.K.");

        case SANE_TYPE_FIXED:
	        if (help) fprintf(stderr,"-%-10.10s          %s\n",
        		opt->name,opt->desc);
        	if ((gopt=getoption(argc,argv,opt->name))==NULL) break;
	    	hlp=(int) (atof(gopt)*(1<<SANE_FIXED_SCALE_SHIFT));
	    	if (verbose) fprintf(stderr,"%d\n",hlp);
	    	if (sane_control_option(scan_handle,x,SANE_ACTION_SET_VALUE,&hlp,NULL)<0) 
		  errout("Option not O.K.");

        case SANE_TYPE_STRING :
        	if  (help) fprintf(stderr,"-%-10.10s          %s\n",
        			opt->name,opt->desc);
        	if ((gopt=getoption(argc,argv,opt->name))==NULL) break;
	    	if (sane_control_option(scan_handle,x,SANE_ACTION_SET_VALUE,gopt,NULL)<0) 
		  errout("Option not O.K.");
		break;

        case SANE_TYPE_BUTTON:
        	if  (help) fprintf(stderr,"-%-10.10s          %s\n",
        			opt->name,opt->desc);
        	if ((gopt=getoption(argc,argv,opt->name))==NULL) break;
	    	if (sane_control_option(scan_handle,x,SANE_ACTION_SET_VALUE,NULL,NULL)<0) 
		  errout("Option not O.K.");
		break;
      }
    }
    if (!help) scan_it();

    sane_close(scan_handle);
  }
  sane_exit();

  return 0;
}

