/* Scan 1.0 -- "image filter" plug-in for the GIMP image manipulation program
 * This one uses the SANE library.
 * Copyright (C) 1996 Andreas Beck
 *
 * You can contact me at becka@sunserver1.rz.uni-duesseldorf.de
 * You can contact the original The GIMP authors at gimp@xcf.berkeley.edu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* Parts of this code have been shamelessly stolen from duplicate.c,
   map_sphere.c and despeckle.c .
*/

/* TODO: show resulting file size.
         have a look at http://www.dcs.ed.ac.uk/home/tst/gimp.html
*/

/* This plug-in acts as an interface for scanners. It should rather be attached
   to "File-Scan", but GIMP doesn't support this now.
*/

#undef DEBUG

#ifdef DEBUG
#define TRACE(a) a
#else
#define TRACE(a)
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "../gimp.h"
#include "sane.h"

#define MAXdlg 128

/* Declare local functions. */

static void scale_callback (int, void *, void *);
static void toggle_callback (int, void *, void *);
static void ok_callback (int, void *, void *);
static void cancel_callback(int, void *, void *);

static char *prog_name;

static SANE_Handle scan_handle;
static int dialog,group,infoID;
static int IDarr[MAXdlg];
static int IDanz;
static SANE_Option_Descriptor *DEarr[MAXdlg];
static volatile int initialselect=0;
static dialog_shown=0;

static int get_opt_num(int item_id)
{ int x;
  for(x=0;x<IDanz;x++) if (IDarr[x]==item_id) return x;
  return -1;
}

static void showhide(int item,int onoff)
{ if (dialog_shown)
  { if (onoff)  gimp_hide_item(dialog,item);
    else        gimp_show_item(dialog,item);
  }
}

static void update_param(void)
{ char buf[200];
  SANE_Parameters sp;
  if (sane_get_parameters(scan_handle,&sp)==SANE_STATUS_GOOD)
    sprintf(buf,"%d x %d : %d kByte",sp.pixels_per_line,
	    sp.lines,(sp.bytes_per_line*sp.lines)>>10);
  else
    sprintf(buf,"Invalid parameters.");
  gimp_change_item(dialog, infoID, strlen(buf)+1 , buf );
  gimp_update_dialog(dialog);
}

static int c_lock=0;

static void update_complete_dialog(void)
{ int x;
  const SANE_Option_Descriptor *opt;
  c_lock=1;
  for(x=0;(opt=sane_get_option_descriptor(scan_handle,x))!=NULL;x++)
  { TRACE(printf("update_complete %d.\n",x));
    switch(opt->type)
    { default:break;
      case SANE_TYPE_BOOL :
	              { SANE_Word tog;
      			sane_control_option(scan_handle,x,SANE_ACTION_GET_VALUE,&tog,NULL);
	      	        gimp_change_item(dialog, IDarr[x], sizeof(tog), &tog );
	      	        showhide(IDarr[x],!SANE_OPTION_IS_ACTIVE(opt->cap));
	    	      } break;
      case SANE_TYPE_INT: 
	              { SANE_Word tog;
      			sane_control_option(scan_handle,x,SANE_ACTION_GET_VALUE,&tog,NULL);
	      	        gimp_change_item(dialog, IDarr[x], sizeof(tog), &tog );
	      	        showhide(IDarr[x],!SANE_OPTION_IS_ACTIVE(opt->cap));
	    	      } break;
      case SANE_TYPE_FIXED:
	              { SANE_Word tog;
      			sane_control_option(scan_handle,x,SANE_ACTION_GET_VALUE,&tog,NULL);
			tog>>=SANE_FIXED_SCALE_SHIFT;
	      	        gimp_change_item(dialog, IDarr[x], sizeof(tog), &tog );
	      	        showhide(IDarr[x],!SANE_OPTION_IS_ACTIVE(opt->cap));
	    	      } break;
      case SANE_TYPE_STRING:
	              { char buf[512];
      			sane_control_option(scan_handle,x,SANE_ACTION_GET_VALUE,buf,NULL);
	      	        gimp_change_item(dialog, IDarr[x], strlen(buf)+1, buf );
	      	        showhide(IDarr[x],!SANE_OPTION_IS_ACTIVE(opt->cap));
		      } break;
    }
  }
  gimp_update_dialog(dialog);
  c_lock=0;
}

static void scale_callback(int item_id, void *client_data, void *call_data)
{ static int lock=0;
  SANE_Word len,info;
  int optnum;
  if (!lock&&!c_lock)
  { lock=1;
    optnum=get_opt_num(item_id);
    switch(DEarr[optnum]->type)
    { case SANE_TYPE_INT:
	len =* (long *)call_data;break;
      case SANE_TYPE_FIXED: 
	len =(*(long *)call_data)<<SANE_FIXED_SCALE_SHIFT;break;
      default: break;
    }
    sane_control_option(scan_handle,optnum,SANE_ACTION_SET_VALUE,&len,&info);
    TRACE(printf("Scale CB info is %x,len=%d.\n",info,len));
    if (info & SANE_INFO_INEXACT)
      { 
	switch(DEarr[optnum]->type)
	  { case SANE_TYPE_INT:
	    default:
	      break;
	    case SANE_TYPE_FIXED: 
	      len>>=SANE_FIXED_SCALE_SHIFT;
	  }
      	gimp_change_item(dialog, item_id, sizeof(len), &len );
      	gimp_update_dialog(dialog);
      }
    if (info & SANE_INFO_RELOAD_OPTIONS) update_complete_dialog();
    if (info & SANE_INFO_RELOAD_PARAMS) update_param();
  }
  lock=0;
}

static void toggle_callback(int item_id, void *client_data, void *call_data)
{ static int lock=0;
  SANE_Word len,info;
  int optnum;
  if (!lock&&!c_lock)
  { lock=1;
    optnum=get_opt_num(item_id);
    len = *(long *)call_data;
    sane_control_option(scan_handle,optnum,SANE_ACTION_SET_VALUE,&len,&info);
    if (info & SANE_INFO_INEXACT)
      { gimp_change_item(dialog, item_id, sizeof(len), &len );
      	gimp_update_dialog(dialog);
      }
    if (info & SANE_INFO_RELOAD_OPTIONS) update_complete_dialog();
    if (info & SANE_INFO_RELOAD_PARAMS) update_param();
  }
  lock=0;
}

static void text_callback(int item_id, void *client_data, void *call_data)
{ static int lock=0;
  SANE_Word info;
  int optnum;
  char buf[512];
  if (!lock&&!c_lock)
  { lock=1;
    optnum=get_opt_num(item_id);
    strcpy(buf,call_data);
    sane_control_option(scan_handle,optnum,SANE_ACTION_SET_VALUE,buf,&info);
    if (info & SANE_INFO_INEXACT)
      { gimp_change_item(dialog, item_id, strlen(buf)+1, buf );
      	gimp_update_dialog(dialog);
      }
    if (info & SANE_INFO_RELOAD_OPTIONS) update_complete_dialog();
    if (info & SANE_INFO_RELOAD_PARAMS) update_param();
  }
  lock=0;
}

static void button_callback(int item_id, void *client_data, void *call_data)
{ static int lock=0;
  SANE_Word info;
  int optnum;
  if (!lock&&!c_lock)
  { lock=1;
    optnum=get_opt_num(item_id);
    sane_control_option(scan_handle,optnum,SANE_ACTION_SET_VALUE,NULL,&info);
    if (info & SANE_INFO_RELOAD_OPTIONS) update_complete_dialog();
    if (info & SANE_INFO_RELOAD_PARAMS) update_param();
  }
  lock=0;
}

static void ok_callback (int item_ID, void *client_data, void *call_data)
{
  SANE_Parameters parm;
  char *d;
  int len,cnt,status;
  Image output=NULL;
  
  do{
  
    status=sane_start(scan_handle);
    if (status!=SANE_STATUS_GOOD) 
    { if (status!=SANE_STATUS_EOF) printf("sane_start failed.\n"); 
      break;
    }

    sane_get_parameters(scan_handle,&parm);
    TRACE(printf("SANE got parameters %dx%d.\n",parm.pixels_per_line,parm.lines));
  
    if (parm.first_frame)	/* Open a new image */
      output = gimp_new_image (0,parm.pixels_per_line,parm.lines,
			       (parm.format==SANE_FRAME_GREY)  ? GRAY_IMAGE : RGB_IMAGE );

    d = gimp_image_data (output);
    gimp_init_progress ("Receiving file ...");cnt=0;
      
    switch(parm.format)
    { case SANE_FRAME_GREY:
      case SANE_FRAME_RGB :
	while(sane_read(scan_handle,d,parm.bytes_per_line,&len)==SANE_STATUS_GOOD)
	{ d+=len;cnt+=len;
	  gimp_do_progress(cnt,parm.bytes_per_line*parm.lines);
	}
	break;
      default:	/* This client does not know how to handle other types of data */
        break;
    }
    gimp_do_progress(1,1);

  } while(1);

  sane_cancel(scan_handle);
  if (output)
  { gimp_display_image (output);
    gimp_update_image (output);
    gimp_free_image (output);
  }
}

static void cancel_callback (int item_ID, void *client_data, void *call_data)
{
  gimp_close_dialog(dialog,0);
}

static void scsel_callback (int item_ID, void *client_data, void *call_data)
{ int x;
  if (!*(long *)call_data) return;
  for(x=0;x<IDanz;x++)
  { if (IDarr[x]==item_ID) { initialselect=x; break; } }
}

int main (int argc, char **argv)
{
  char buf[1000];
  const SANE_Option_Descriptor *opt;
  int x;
  int lastgroupID,DummyID;
  const SANE_Device **devlist;
  SANE_Status status;
  char *selection;

  /* Save the program name so we can use it later in reporting errors
   */
  prog_name = argv[0];

  /* Call 'gimp_init' to initialize this filter.
   * 'gimp_init' makes sure that the filter was properly called and
   *  it opens pipes for reading and writing.
   */
  if (gimp_init (argc, argv))
    {
      status=sane_init(NULL);
      TRACE(printf("SANE init.%s.\n",sane_strstatus(status)));

      /* Build Scanner Select list. */
      sane_get_devices(&devlist);
      TRACE(printf("SANE got devlist %s.%s.\n",devlist[0]->name,sane_strstatus(status)));

      /**/ if (devlist[0]==NULL) selection=""; /* No devices ... bail out */
      else if (devlist[1]==NULL) selection=devlist[initialselect=0]->name;
      else { dialog=gimp_new_dialog("Select Scanner");
             group =gimp_new_row_group(dialog, DEFAULT, RADIO, "");
             for(x=0;devlist[x]!=NULL;x++)
             { IDarr[x]=gimp_new_radio_button (dialog, group, devlist[x]->name);
               gimp_add_callback ( dialog, IDarr[x], scsel_callback, &IDarr[x]);
#ifdef DEBUG
		sprintf(buf,"Vendor: %s",devlist[x]->vendor);
		gimp_new_label(dialog, group, buf);
	        sprintf(buf,"Model: %s",devlist[x]->model);
		gimp_new_label(dialog, group, buf);
	        sprintf(buf,"Type:%s",devlist[x]->type);
		gimp_new_label(dialog, group, buf);
#endif
             }
             IDanz=x;
	     gimp_add_callback(dialog,gimp_ok_item_id (dialog) , cancel_callback, 0);
	     gimp_add_callback(dialog,gimp_cancel_item_id (dialog) , cancel_callback, 0);
	     gimp_show_dialog(dialog);
	     selection=devlist[initialselect]->name;
	   }

      if ((status=sane_open(selection,&scan_handle))==SANE_STATUS_GOOD)
      { /* This is NOT a regular filter. It needs no input image. */

        TRACE(printf("SANE opened %s, handle =%p.\n",selection,scan_handle));
        dialog= gimp_new_dialog ("Scanner Control");
        group = gimp_new_row_group (dialog,DEFAULT,NORMAL,"");
#ifdef DEBUG
        sprintf(buf,"Vendor: %s",devlist[initialselect]->vendor);
	gimp_new_label(dialog, group, buf);
        sprintf(buf,"Model: %s",devlist[initialselect]->model);
	gimp_new_label(dialog, group, buf);
        sprintf(buf,"Type:%s",devlist[initialselect]->type);
	gimp_new_label(dialog, group, buf);
#endif
        sprintf(buf,"Name:%s",devlist[initialselect]->name);
	gimp_new_label(dialog, group, buf);
	for(x=0;(opt=sane_get_option_descriptor(scan_handle,x))!=NULL;x++)
	{ DEarr[x]=opt;
          TRACE(printf("SANE get_opt_desc %d.\n",x));
          switch(opt->type)
	  { 
	    case SANE_TYPE_GROUP:
	      IDarr[x]=gimp_new_frame (dialog, group, opt->name);
	      lastgroupID=gimp_new_row_group(dialog, IDarr[x], NORMAL, "");
              TRACE(printf("SANE made group %s.\n",opt->name));
	      break;
	    case SANE_TYPE_BOOL: /* We use a checkbutton for bool */
	      IDarr[x]=gimp_new_check_button(dialog, lastgroupID, opt->name);
	      gimp_add_callback(dialog,IDarr[x],toggle_callback, NULL);
	      { SANE_Word onoff;
	        sane_control_option(scan_handle,x,SANE_ACTION_GET_VALUE,&onoff,NULL);
      	        gimp_change_item(dialog, IDarr[x], sizeof(onoff), &onoff );
	      }
              TRACE(printf("SANE made toggle %s.\n",opt->name));
	      break;
	    case SANE_TYPE_INT:
	    case SANE_TYPE_FIXED:/* Scales for these */
	      { SANE_Word sca;
		switch(opt->constraint_type)
		  { case SANE_CONSTRAINT_RANGE: /* Use a scale */
		      DummyID = gimp_new_column_group(dialog,lastgroupID, NORMAL, "");
		      gimp_new_label(dialog, DummyID, opt->name);
		      sane_control_option(scan_handle,x,SANE_ACTION_GET_VALUE,&sca,NULL);
		      switch(opt->type)
			{ case SANE_TYPE_INT:
			    IDarr[x]= gimp_new_scale (dialog,DummyID, 
						      opt->constraint.range->min, 
						      opt->constraint.range->max, 
						      sca, 0);
			    break;
			  case SANE_TYPE_FIXED:
			    IDarr[x]= gimp_new_scale (dialog,DummyID, 
						      opt->constraint.range->min>>SANE_FIXED_SCALE_SHIFT, 
						      opt->constraint.range->max>>SANE_FIXED_SCALE_SHIFT, 
						      sca>>SANE_FIXED_SCALE_SHIFT
						      , 0);
			    break;
			  default: TRACE(printf("This can't happen.\n"));break;
			}
		      TRACE(printf("SANE made scale %s.\n",opt->name));
		      gimp_add_callback(dialog,IDarr[x],scale_callback, NULL);
		      break;
		    case SANE_CONSTRAINT_WORD_LIST: /* Use a "list-select" widget */
		    case SANE_CONSTRAINT_NONE:/* Use a text widget*/
		    default:break;
		  }
	      } break;
	    case SANE_TYPE_STRING : 
	      { DummyID = gimp_new_column_group(dialog,lastgroupID, NORMAL, "");
		gimp_new_label(dialog, DummyID, opt->name);
		sane_control_option(scan_handle,x,SANE_ACTION_GET_VALUE,buf,NULL);
		IDarr[x]= gimp_new_text (dialog,DummyID,buf);
                TRACE(printf("SANE made text %s,%s.\n",opt->name,buf));
		gimp_add_callback(dialog,IDarr[x],text_callback, NULL);
	      } break;
	    case SANE_TYPE_BUTTON : 
	      IDarr[x] = gimp_new_push_button (dialog, lastgroupID, opt->name);
	      gimp_add_callback(dialog,IDarr[x], button_callback, &IDarr[x]);
	      break;
	    default:IDarr[x]=-1;break;
	  }
	}
	IDanz=x;

        sprintf(buf,"INFO:");
	infoID=gimp_new_text(dialog, group, buf);

        gimp_add_callback(dialog,gimp_ok_item_id (dialog) , ok_callback, 0);
        gimp_add_callback(dialog,gimp_cancel_item_id (dialog) , cancel_callback, 0);

	update_param();
#if 0
	update_complete_dialog();
#endif
        dialog_shown=1;
        gimp_show_dialog(dialog);

        sane_close(scan_handle);
      }
      TRACE(printf("Closing sane status %s.\n",sane_strstatus(status)));
      sane_exit();
    }

  /* Quit
   */

  gimp_quit ();

  return 0;
}

