#include "sane.h"
#include "saneopts.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>

#define V_MAJOR 0
#define V_MINOR 0
#define V_BUILD 1
#define MAGIC   (void *)0xab730324

#if V_MAJOR != SANE_CURRENT_MAJOR
#error Major has changed in sane.h ! Check this code !
#endif

/*\
|*| This library is a demo implementation of a SANE backend.
|*| It implements a virtual device, a PBM file-filter.
\*/

SANE_Status sane_init (SANE_Word * version_code)
{ 
  if (version_code) *version_code=SANE_VERSION_CODE(V_MAJOR,V_MINOR,V_BUILD);
  return SANE_STATUS_GOOD;
}

SANE_Status sane_exit (void)
{ 
  return SANE_STATUS_GOOD; 
}

/****************************************************************/
/* Device select/open/close                                     */
/****************************************************************/

static const SANE_Device pbmdev=
{ "PBM virtual file driver",
  "None",
  "PBM file driver",
  "virtual device"
};

static const SANE_Device pbmdev2=
{ "2nd PBM file driver",
  "None",
  "PBM file driver",
  "virtual device"
};

static SANE_Device const *dev_list[]=
{ &pbmdev,
  &pbmdev2,
  NULL
};

SANE_Status sane_get_devices (const SANE_Device * * * device_list)
{
  *device_list=dev_list;
  return SANE_STATUS_GOOD; 
}

static int is_open=0;

SANE_Status sane_open (SANE_String devicename, SANE_Handle * handle)
{ 
  if (strcmp(devicename,pbmdev.name)==0||
      strcmp(devicename,pbmdev2.name)==0) 
    { if (is_open) return SANE_STATUS_DEVICE_BUSY;
      else { is_open=1;*handle=MAGIC;return SANE_STATUS_GOOD; }	
      /* Single-access driver */
    }
  return SANE_STATUS_INVAL;
}

extern SANE_Status sane_close (SANE_Handle handle)
{ 
  if (handle!=MAGIC) return SANE_STATUS_INVAL; /* Unknown handle ... */ 
  is_open=0;
  return SANE_STATUS_GOOD;
}

char filename[128]="/var/.graph/tmp.ppm";

SANE_Word bright= 0;
SANE_Word contr = 0;
SANE_Word grey  = 0;


static const SANE_Option_Descriptor sod[]=
{ 
  { SANE_OPT_NUM_OPTIONS,
    SANE_DESC_NUM_OPTIONS,
    SANE_TYPE_INT,
    SANE_UNIT_NONE,
    sizeof(SANE_Word),
    SANE_CAP_SOFT_DETECT,
    SANE_CONSTRAINT_NONE,
    { NULL }    
  },
  { "File:",
    "Selection of the file to load.",
    SANE_TYPE_GROUP,
    SANE_UNIT_NONE,
    0,
    SANE_CAP_SOFT_DETECT,
    SANE_CONSTRAINT_NONE,
    { NULL }
  },
  { SANE_OPT_FILE,
    SANE_DESC_FILE,
    SANE_TYPE_STRING,
    SANE_UNIT_NONE,
    sizeof(filename),
    SANE_CAP_SOFT_SELECT|SANE_CAP_SOFT_DETECT,
    SANE_CONSTRAINT_NONE,
    { NULL }
  },
  { "Image Enhancement:",
    "A few controls to enhance image while loading:",
    SANE_TYPE_GROUP,
    SANE_UNIT_NONE,
    0,
    SANE_CAP_SOFT_DETECT,
    SANE_CONSTRAINT_NONE,
    { NULL }
  },
  { SANE_OPT_BRIGHTNESS,
    SANE_DESC_BRIGHTNESS,
    SANE_TYPE_FIXED,
    SANE_UNIT_PERCENT,
    sizeof(SANE_Word),
    SANE_CAP_SOFT_SELECT|SANE_CAP_SOFT_DETECT,
    SANE_CONSTRAINT_RANGE,
    { &percentage_range }
  },
  { SANE_OPT_CONTRAST,
    SANE_DESC_CONTRAST,
    SANE_TYPE_FIXED,
    SANE_UNIT_PERCENT,
    sizeof(SANE_Word),
    SANE_CAP_SOFT_SELECT|SANE_CAP_SOFT_DETECT,
    SANE_CONSTRAINT_RANGE,
    { &percentage_range }
  },
  { "Grayify",
    "Load the image as greyscale.",
    SANE_TYPE_BOOL,
    SANE_UNIT_NONE,
    sizeof(SANE_Word),
    SANE_CAP_SOFT_SELECT|SANE_CAP_SOFT_DETECT,
    SANE_CONSTRAINT_NONE,
    { NULL }
  },
  { "Defaults",
    "Load defaults for enhancement.",
    SANE_TYPE_BUTTON,
    SANE_UNIT_NONE,
    0,
    SANE_CAP_SOFT_SELECT|SANE_CAP_SOFT_DETECT,
    SANE_CONSTRAINT_NONE,
    { NULL }
  }
};

#define SANE_OptAnz (sizeof(sod)/sizeof(SANE_Option_Descriptor))

const SANE_Option_Descriptor * sane_get_option_descriptor (SANE_Handle handle,
                                                            SANE_Word option)
{
  if (handle!=MAGIC || !is_open) return NULL; /* wrong device */
  if (option<0||option>=SANE_OptAnz)
    return NULL;
  return &sod[option];
}

extern SANE_Status sane_control_option (SANE_Handle handle, SANE_Word option,
                                        SANE_Action action, void * value,
                                        SANE_Word * info)
{ 
  SANE_Word myinfo=0;
  if (handle!=MAGIC || !is_open) return SANE_STATUS_INVAL; /* Unknown handle ... */ 
  if (option<0||option>=SANE_OptAnz) return SANE_STATUS_INVAL; /* Unknown option ... */ 
  switch(action)  
    { case SANE_ACTION_SET_VALUE:switch(option)
      	{ case 2: strcpy(filename,value);
      		  if (access(filename,R_OK)==0)
      		    myinfo|=SANE_INFO_RELOAD_PARAMS;
      		  break;
          case 4: bright=*(SANE_Word *)value;
          	  break;
          case 5: contr=*(SANE_Word *)value;
          	  break;
       	  case 6: grey=!!*(SANE_Word *)value;
       	  	  myinfo|=SANE_INFO_RELOAD_PARAMS;
       	  	  break;
       	  case 7: bright=contr=0;
       	          myinfo|=SANE_INFO_RELOAD_OPTIONS;
       	          break;
      	  default: return SANE_STATUS_INVAL;
      	}
      case SANE_ACTION_GET_VALUE:switch(option)
      	{ case 0: *(SANE_Word *)value=SANE_OptAnz;break; 
      	  case 2: strcpy(value,filename);break;
          case 4: *(SANE_Word *)value=bright;break;
          case 5: *(SANE_Word *)value=contr;break;
       	  case 6: *(SANE_Word *)value=grey;break;
       	  case 7: break;
      	  default: return SANE_STATUS_INVAL;
      	}
      	break;
      case SANE_ACTION_SET_AUTO: return SANE_STATUS_INVAL; /* We are DUMB */
      ;
    }
  if (info) *info=myinfo;
  return SANE_STATUS_GOOD;
}

static void getline(char *buf,int len,FILE *f)
{ do{ fgets(buf,len,f); }while(*buf=='#'); }

SANE_Parameters parms=
{ SANE_FRAME_RGB,
  1, /* Always first frame, no multichannel */
  0, /* Number of bytes returned per scan line: */
  0, /* Number of pixels per scan line.  */
  0, /* Number of lines for the current scan.  */
  8, /* Number of bits per sample. */
};
                                                                            
static int getparmfromfile(void)
{ FILE *fn;
  int x,y;
  char buf[1024];

  parms.bytes_per_line=parms.pixels_per_line=parms.lines=0;
  if ((fn=fopen(filename,"r"))==NULL) return -1;

  getline(buf,sizeof(buf),fn);
  if (strncmp(buf,"P6",2)) { fclose(fn);return -1; }
  getline(buf,sizeof(buf),fn);
  sscanf(buf,"%d %d",&x,&y);

  parms.bytes_per_line=x;
  parms.pixels_per_line=x;
  parms.lines=y;
  if (grey) parms.format=SANE_FRAME_GREY; 
  else    { parms.format=SANE_FRAME_RGB;parms.bytes_per_line*=3; }

  return 0;
}

extern SANE_Status sane_get_parameters (SANE_Handle handle,
					SANE_Parameters * params)
{
  int rc=SANE_STATUS_GOOD;
  if (handle!=MAGIC || !is_open) rc=SANE_STATUS_INVAL; /* Unknown handle ... */ 
  else if (getparmfromfile()) rc=SANE_STATUS_INVAL;
  *params=parms;
  return rc;
}

FILE *infile=NULL;

extern SANE_Status sane_start (SANE_Handle handle)
{
  char buf[1024];
  if (handle!=MAGIC || !is_open) return SANE_STATUS_INVAL; /* Unknown handle ... */ 
  if (getparmfromfile()) return SANE_STATUS_INVAL;
  if (infile!=NULL) { fclose(infile);infile=NULL;return SANE_STATUS_EOF; }
  if ((infile=fopen(filename,"r"))==NULL) return SANE_STATUS_INVAL;
  getline(buf,sizeof(buf),infile);/* Skip the header */
  getline(buf,sizeof(buf),infile);
  getline(buf,sizeof(buf),infile);/* Can there be comments after the last header ??? */
  
  return SANE_STATUS_GOOD;
}

extern SANE_Status sane_read (SANE_Handle handle, SANE_Byte * data,
			      SANE_Word max_length, SANE_Word * length)
{
  int len,x,hlp;
  if (handle!=MAGIC || !is_open) return SANE_STATUS_INVAL; /* Unknown handle ... */ 
  if (!grey)
  { len=fread(data,1,max_length,infile);
  } else
  { unsigned char buf[3];
    for(len=0;len<max_length;len++)
    { if (fread(buf,1,sizeof(buf),infile)!=sizeof(buf)) break;
      *(data+len)=((long)buf[0]+buf[1]+buf[2])/3;
    }
  }

  /* Do the transformations ... DEMO ONLY ! THIS MAKES NO SENSE ! */
  for(x=0;x<len;x++)
  { hlp=*((unsigned char *)data+x)-128;
    hlp*=(contr+(100<<SANE_FIXED_SCALE_SHIFT));
    hlp/=100<<SANE_FIXED_SCALE_SHIFT;
    hlp+=(bright>>SANE_FIXED_SCALE_SHIFT)+128;
    if (hlp<0) hlp=0;if (hlp>255) hlp=255;
    *(data+x)=hlp;      
  }

  if (len!=max_length) return SANE_STATUS_EOF;
  *length=len;
  return SANE_STATUS_GOOD;
}

extern SANE_Status sane_cancel (SANE_Handle handle)
{
  if (infile!=NULL) { fclose(infile);infile=NULL; }
  return SANE_STATUS_GOOD;
}

extern SANE_Status sane_set_io_mode (SANE_Handle handle,
				     SANE_Word non_blocking)
{
  return SANE_STATUS_UNSUPPORTED;
}

extern SANE_Status sane_get_select_fd (SANE_Handle handle,
				       SANE_Word * fd)
{
  return SANE_STATUS_UNSUPPORTED;
}

extern const SANE_String sane_strstatus (SANE_Status status)
{
  switch (status)
  { case SANE_STATUS_GOOD: return "O.K. - no error";
    case SANE_STATUS_UNSUPPORTED: return "Operation not supported";
    case SANE_STATUS_DEVICE_BUSY: return "Device is Busy";
    case SANE_STATUS_AGAIN: return "No data yet - please retry";
    case SANE_STATUS_INVAL: return "Invalid data specified";
    case SANE_STATUS_EOF: return "End of file reached";
    default: return "unknown SANE status code";
  }
}
